@@ -412,9 +412,11 @@ grpnet <- function(
     # compute common quantities
     if (!is.null(offsets)) {
         offsets <- as.double(offsets)
-        if ((dim(offsets) != dim(y)) ||
+        dim.offsets <- dim(offsets)
+        dim.y <- dim(y)
+        if (((!is.null(dim.offsets) || !is.null(dim.y)) && !identical(dim.offsets, dim.y)) ||
             (length(offsets) != length(y))
-        ) {
+            ) {
             stop("offsets must be same shape as y if not NULL.")
         }
     } else {
