@@ -39,7 +39,7 @@ private:
     const size_t _K;
     const size_t _n_threads;
     vec_value_t _buff;
-    
+
 public:
     explicit MatrixNaiveKroneckerEye(
         base_t& mat,
@@ -48,17 +48,7 @@ public:
     );
 
     ADELIE_CORE_MATRIX_NAIVE_PURE_OVERRIDE_DECL
-
-    virtual void mean(
-        const Eigen::Ref<const vec_value_t>& weights,
-        Eigen::Ref<vec_value_t> out
-    ) override;
-
-    virtual void var(
-        const Eigen::Ref<const vec_value_t>& centers,
-        const Eigen::Ref<const vec_value_t>& weights,
-        Eigen::Ref<vec_value_t> out
-    ) override;
+    ADELIE_CORE_MATRIX_NAIVE_OVERRIDE_DECL
 };
 
 template <class DenseType, class IndexType=Eigen::Index> 
@@ -80,7 +70,23 @@ private:
     const size_t _n_threads;
     rowmat_value_t _buff;
     vec_value_t _vbuff;
-    
+
+    inline auto _cmul(
+        int j, 
+        const Eigen::Ref<const vec_value_t>& v,
+        const Eigen::Ref<const vec_value_t>& weights,
+        Eigen::Ref<rowmat_value_t> buff
+    ) const;
+
+    inline void _bmul(
+        int j, int q, 
+        const Eigen::Ref<const vec_value_t>& v, 
+        const Eigen::Ref<const vec_value_t>& weights,
+        Eigen::Ref<vec_value_t> out,
+        Eigen::Ref<vec_value_t> vbuff,
+        Eigen::Ref<rowmat_value_t> buff
+    ) const;
+
 public:
     explicit MatrixNaiveKroneckerEyeDense(
         const Eigen::Ref<const dense_t>& mat,
@@ -89,17 +95,7 @@ public:
     );
 
     ADELIE_CORE_MATRIX_NAIVE_PURE_OVERRIDE_DECL
-
-    virtual void mean(
-        const Eigen::Ref<const vec_value_t>& weights,
-        Eigen::Ref<vec_value_t> out
-    ) override;
-
-    virtual void var(
-        const Eigen::Ref<const vec_value_t>& centers,
-        const Eigen::Ref<const vec_value_t>& weights,
-        Eigen::Ref<vec_value_t> out
-    ) override;
+    ADELIE_CORE_MATRIX_NAIVE_OVERRIDE_DECL
 };
 
 } // namespace matrix
